/*
 * Pad for Psemu Pro like Emulators
 *
 * By ammoQ <ammoq@ammoq.com>
 * Based on padXwin by linuzappz <linuzappz@hotmail.com>
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/joystick.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

typedef char HWND;
#include "PSEmu Plugin Defs.h"

char *LibName = "ammoQ Joydevice Pad Driver";

const unsigned char version = 1; // PSEmu 1.x library
const unsigned char revision = VERSION;
const unsigned char build = BUILD;


// Prototypes
static void loadConfig();
static void saveConfig();

// Filenames for device files, e.g. "/dev/input/js0"
static char devicefilename[2][FILENAME_MAX+1] = {"/dev/input/js0", ""};

// File desciptors for device files
static int devicefile[2] = { -1, -1 };

// use PCSX return values?
static int pcsx_style = 0;

// Position of directional buttons in PadButtons
#define PSXBTN_UP 12
#define PSXBTN_RIGHT 13
#define PSXBTN_DOWN 14
#define PSXBTN_LEFT 15

// Zero range (for analogue sticks)
#define MINZERO -50
#define MAXZERO 50

// Assignment of PSX buttons to device buttons
static short PadButtons[2][16] = {{
    10,	// L2
    11, // R2
    6,  // L1
    7,	// R1
    4,	// Triangle
    1,	// Circle
    0,	// Cross
    3,	// Square
    9,  // Select
    -1,	//
    -1,	// 
     8, // Start
    -1,	// Up
    -1,	// Right
    -1,	// Down
    -1	// Left
},
{ -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1 }};

static int PadOpened = 0;
static int PadFlags = 0;
static Display *Dsp;

char *PSEgetLibName(void) {
    return LibName;
}

unsigned long PSEgetLibType(void) {
    return PSE_LT_PAD;
}

unsigned long PSEgetLibVersion(void) {
    return version<<16|revision<<8|build;
}

long PADinit(long flags) {
    PadFlags|=flags;
    loadConfig();

    return 0;
}

long PADshutdown(void) {
    return 0;
}

long PADopen(unsigned long *Disp) {
    int i;
    Dsp = (Display *)*Disp;
    PadOpened = 1;

    for (i=0; i<2; i++) {
        if (devicefilename[i][0]) {
            devicefile[i] = open(devicefilename[i], O_RDONLY);
            if (devicefile[i] == -1) {
                fprintf(stderr, "padJoy: could not open %s\n", devicefilename[i]);
            }
        }
    }

    return 0;
}

long PADclose(void) {
    int i;
    for (i=0; i<2; i++) {
        if (devicefile[i] > -1) {
            close (devicefile[i]);
        }
    }

    PadOpened = 0;
    return 0;
}

long PADquery(void) {
    return 3; // both pads
}

typedef struct
{
    unsigned char controllerType;
    unsigned short buttonStatus;
    unsigned char rightJoyX, rightJoyY, leftJoyX, leftJoyY;
    unsigned char moveX, moveY;
    unsigned char reserved[91];
} PadDataS;


short PressedKey;
short ReleasedKey;
unsigned short PadStat[2] = {0xffff, 0xffff};

static void CheckPad(int pad) {
    fd_set rfds;
    struct timeval tv;
    int retval;
    struct js_event e;
    int i;

    if (devicefile[pad] >-1) {
        do {
            // Test if there is input waiting for us...
            FD_ZERO(&rfds);
            FD_SET(devicefile[pad], &rfds);
            tv.tv_sec = 0;
            tv.tv_usec = 0;

            retval = select(devicefile[pad]+1, &rfds, NULL, NULL, &tv);
            if (retval) {
                read (devicefile[pad], &e, 8);

                if (e.type == JS_EVENT_AXIS) {
                    // this is the stick
                    if (e.number == 0) {
                        // left-right axis
                        if (e.value > MAXZERO) {
                            // right
                            PadStat[pad]&=~(1<<PSXBTN_RIGHT);
                            PadStat[pad]|=(1<<PSXBTN_LEFT); 
                        }
                        else if (e.value < MINZERO) {
                            // left
                            PadStat[pad]&=~(1<<PSXBTN_LEFT);
                            PadStat[pad]|=(1<<PSXBTN_RIGHT); 
                        }
                        else {
                            PadStat[pad]|=(1<<PSXBTN_LEFT); 
                            PadStat[pad]|=(1<<PSXBTN_RIGHT); 
                        }
                    }
                    else if (e.number == 1) {
                        // up-down axis
                        if (e.value > MAXZERO) {
                            // down
                            PadStat[pad]&=~(1<<PSXBTN_DOWN);
                            PadStat[pad]|=(1<<PSXBTN_UP); 
                        }
                        else if (e.value < MINZERO) {
                            // up
                            PadStat[pad]&=~(1<<PSXBTN_UP);
                            PadStat[pad]|=(1<<PSXBTN_DOWN); 
                        }
                        else {
                             PadStat[pad]|=(1<<PSXBTN_UP); 
                             PadStat[pad]|=(1<<PSXBTN_DOWN); 
                        }
                    }
                     // other axis could be mapped to analoge stick...
                }
                else if (e.type == JS_EVENT_BUTTON) {
                    for (i=0; i<16; i++) {
                        if (PadButtons[pad][i] == e.number) {
                            if (e.value)
                                PadStat[pad]&=~(1<<i);
                            else
                                PadStat[pad]|=(1<<i); 
                        }
                    }
                }
            } // if retval
        } while (retval);
    }  // if devicefile[pad] open
}
    

long PADreadPort1(PadDataS *pad) {
    CheckPad(0);

    pad->buttonStatus = PadStat[0];
    pad->controllerType = 4; // standard

    if (pcsx_style) { // ePSXe different from pcsx, swap bytes 
       pad->buttonStatus = (pad->buttonStatus>>8)|(pad->buttonStatus<<8);
    }

    return 0;
}

long PADreadPort2(PadDataS *pad) {
    CheckPad(1);

    pad->buttonStatus = PadStat[1];
    pad->controllerType = 4; // standard

    if (pcsx_style) { // ePSXe different from pcsx, swap bytes 
       pad->buttonStatus = (pad->buttonStatus>>8)|(pad->buttonStatus<<8);
    }

    return 0;
}

long PADkeypressed(void) {
    return 0;
}

/*---------------------------------------------------------------------*/
/*                          Config Dialogue stuff                      */
/*---------------------------------------------------------------------*/

static void saveConfig() {
    FILE *f;
    int i;

    f=fopen("cfg/padAmmoq.cfg", "w");
    if (!f) {
        fprintf(stderr, "cannot open cfg/padAmmoq.cfg");
        return;
    }

    fprintf(f,"[general]\n");
    fprintf(f,"pcsx_style      = %d\n", pcsx_style);
    for (i=0;i<2;i++) {
        fprintf(f,"[pad %d]\n", i+1);
        fprintf(f,"devicefilename  = %s\n", devicefilename[i]);
        fprintf(f,"button_l2       = %d\n", PadButtons[i][0]);
        fprintf(f,"button_r2       = %d\n", PadButtons[i][1]);
        fprintf(f,"button_l1       = %d\n", PadButtons[i][2]);
        fprintf(f,"button_r1       = %d\n", PadButtons[i][3]);
        fprintf(f,"button_triangle = %d\n", PadButtons[i][4]);
        fprintf(f,"button_circle   = %d\n", PadButtons[i][5]);
        fprintf(f,"button_cross    = %d\n", PadButtons[i][6]);
        fprintf(f,"button_square   = %d\n", PadButtons[i][7]);
        fprintf(f,"button_select   = %d\n", PadButtons[i][8]);
        fprintf(f,"button_start    = %d\n", PadButtons[i][11]);
   }

   fclose(f);
}

static void loadConfig() {
    FILE *f;
    int i;
    char line[FILENAME_MAX+30];
    int pad=0;
    char *val;

    f=fopen("cfg/padAmmoq.cfg", "r");
    if (!f) {
        fprintf(stderr, "cannot open cfg/padAmmoq.cfg");
        return;
    }

    while(!feof(f)) {
        fgets(line, FILENAME_MAX+29, f);
        i=strlen(line)-1;
        while (i>0 && line[i]<32) line[i--]='\0';

        val=NULL;
        while(i>0) {
           if (line[i]=='=') val = line+(i+1);
           i--;
        }
        if (val) {
            while (*val==' ') val++;
        }

        if (!strcmp(line, "[general]")) {
            // nothing to do
        }
        else if (!strncmp(line, "pcsx_style", 10)) {
           pcsx_style = atoi(val);
        }
        else if (!strcmp(line, "[pad 1]")) {
            pad = 0;
        }
        else if (!strcmp(line, "[pad 2]")) {
            pad = 1;
        }
        else if (!strncmp(line, "devicefilename", 14)) {
           strcpy(devicefilename[pad], val);
        }
        else if (!strncmp(line, "button_l2", 9)) PadButtons[pad][0] = atoi(val);
        else if (!strncmp(line, "button_r2", 9)) PadButtons[pad][1] = atoi(val);
        else if (!strncmp(line, "button_l1", 9)) PadButtons[pad][2] = atoi(val);
        else if (!strncmp(line, "button_r1", 9)) PadButtons[pad][3] = atoi(val);
        else if (!strncmp(line, "button_triangle", 15)) PadButtons[pad][4] = atoi(val);
        else if (!strncmp(line, "button_circle", 13)) PadButtons[pad][5] = atoi(val);
        else if (!strncmp(line, "button_cross", 12)) PadButtons[pad][6] = atoi(val);
        else if (!strncmp(line, "button_square", 13)) PadButtons[pad][7] = atoi(val);
        else if (!strncmp(line, "button_select", 13)) PadButtons[pad][8] = atoi(val);
        else if (!strncmp(line, "button_start", 12)) PadButtons[pad][11] = atoi(val);
        else fprintf(stderr, "dont understand %s\n", line);
   }   
}




/* number of buttons required for configuration */
#define CONFIGBUTTONCOUNT 10

static int currentPad=0;

static struct {
  GtkWidget *config_window;

  GSList *group;
  GtkWidget *padno_radio[2];
  GtkWidget *filename_entry;
  GtkWidget *button[CONFIGBUTTONCOUNT];
  GtkWidget *label[CONFIGBUTTONCOUNT];
  GtkWidget *pcsx_check;

  GtkWidget *ok_button;
  GtkWidget *cancel_button;
} ConfWidgets;

static struct { int nr;
         char *label;
         int x;
         int y; }
  buttonInfo[CONFIGBUTTONCOUNT] =
{
{0, "  L2  ", 20, 70},
{2, "  L1  ", 20, 100},
{1, "  R2  ", 350, 70},
{3, "  R1  ", 350, 100},
{4, "  /\\  ", 300, 150},
{7, "  [_]  ", 250, 180},
{5, "  (_)  ", 350, 180},
{6, "  ><  ", 300, 210},
{8, "  Select  ", 100, 210},
{11, "  Start  ", 190, 210}
};

static void showPadConfiguration() {
    int i;
    char buffer[100];

    // gtk_toggle_button_set_active(ConfWidgets.padno_radio[currentPad], TRUE);
    // gtk_toggle_button_set_active(ConfWidgets.padno_radio[1-currentPad], FALSE);
    
    gtk_entry_set_text ( GTK_ENTRY(ConfWidgets.filename_entry), devicefilename[currentPad] );

    for (i=0; i<CONFIGBUTTONCOUNT; i++) {
        sprintf(buffer, "%d",  PadButtons[currentPad][buttonInfo[i].nr]);
        gtk_label_set(GTK_LABEL(ConfWidgets.label[i]), buffer);
    }
}

static void OnConfCancel() {
    loadConfig();
    gtk_widget_hide(ConfWidgets.config_window);
    gtk_widget_destroy(ConfWidgets.config_window);
    gtk_main_quit();
}

static void OnConfOk() {
    saveConfig();
    gtk_widget_hide(ConfWidgets.config_window);
    gtk_widget_destroy(ConfWidgets.config_window);
    gtk_main_quit();
}

static void OnConfBtn(GtkWidget *But, gpointer data) {
    int f;
    fd_set rfds;
    struct timeval tv;
    int retval;
    struct js_event e;
    char buffer[100];

    int labelnr = (int) gtk_object_get_user_data(GTK_OBJECT(But));
    int btnnr = buttonInfo[labelnr].nr;

    f = open(devicefilename[currentPad], O_RDONLY);
    if (f >-1) {
        do {
            // Test if there is input waiting for us...
            FD_ZERO(&rfds);
            FD_SET(f, &rfds);
            tv.tv_sec = 3;  // timeout for configuration...
            tv.tv_usec = 0;

            retval = select(f+1, &rfds, NULL, NULL, &tv);
            if (retval) {
                read (f, &e, 8);

                if (e.type == JS_EVENT_AXIS) {
                     // currently no configuration for it, maybe later if someone needs it...
                }
                else if (e.type == JS_EVENT_BUTTON && e.value) {
                    PadButtons[currentPad][btnnr] = e.number;
                    sprintf(buffer, "%d", e.number);
                    gtk_label_set(GTK_LABEL(ConfWidgets.label[labelnr]), buffer);
                    retval = 0;
                }
            } // if retval
        } while (retval);
        close(f);
        return;
    }  // if f open
}

static void OnPcsxCheck(GtkWidget *But, gpointer data) {
     pcsx_style = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(ConfWidgets.pcsx_check));
}

static void OnPadnoRadio(GtkWidget *But, gpointer data) {
     if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(ConfWidgets.padno_radio[1]))) {
         currentPad = 1;
     }
     else {
         currentPad = 0;
     }

     showPadConfiguration();
}

static void OnFilenameEntry(GtkWidget *Ent, gpointer data) {
    strcpy(devicefilename[currentPad], gtk_entry_get_text(GTK_ENTRY(Ent)));
}

static void CreateConfigWindow(void) {
    char buffer[100];
    int i;
    GtkWidget *hbox;
    GtkWidget *fixed;

    ConfWidgets.config_window = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_widget_set_usize(ConfWidgets.config_window, 460, 320);
    gtk_window_set_title(GTK_WINDOW(ConfWidgets.config_window), "PAD Config Dialog");
    gtk_window_set_position(GTK_WINDOW(ConfWidgets.config_window), GTK_WIN_POS_CENTER);
    gtk_container_set_border_width(GTK_CONTAINER(ConfWidgets.config_window), 10);
    gtk_signal_connect(GTK_OBJECT(ConfWidgets.config_window), "destroy-event", GTK_SIGNAL_FUNC(OnConfCancel), NULL);

    fixed = gtk_fixed_new();
    gtk_container_add(GTK_CONTAINER(ConfWidgets.config_window), fixed);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_fixed_put(GTK_FIXED(fixed), hbox, 10, 10);
    gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("Pad number:"), FALSE, TRUE, 2);

    ConfWidgets.group = NULL;

    for (i=0; i<2; i++) {
        sprintf(buffer, "%d", i+1);
        ConfWidgets.padno_radio[i] = gtk_radio_button_new_with_label ( ConfWidgets.group, buffer);
        gtk_box_pack_start ( GTK_BOX (hbox), ConfWidgets.padno_radio[i], FALSE, FALSE, 0); 
        ConfWidgets.group = gtk_radio_button_group ( GTK_RADIO_BUTTON (ConfWidgets.padno_radio[i]));
        gtk_signal_connect(GTK_OBJECT(ConfWidgets.padno_radio[i]), "toggled", GTK_SIGNAL_FUNC(OnPadnoRadio), NULL);
    }

    ConfWidgets.pcsx_check = gtk_check_button_new_with_label("PCSX");
    gtk_fixed_put(GTK_FIXED(fixed), ConfWidgets.pcsx_check, 250, 10);
     gtk_signal_connect(GTK_OBJECT(ConfWidgets.pcsx_check), "toggled", GTK_SIGNAL_FUNC(OnPcsxCheck), NULL);
    gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(ConfWidgets.pcsx_check), pcsx_style);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_fixed_put(GTK_FIXED(fixed), hbox, 10, 40);
    gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new("Device file:"), FALSE, TRUE, 2);

    ConfWidgets.filename_entry = gtk_entry_new();
    gtk_entry_set_text ( GTK_ENTRY(ConfWidgets.filename_entry), (gchar *) devicefilename[0] );
    gtk_entry_set_max_length ( GTK_ENTRY(ConfWidgets.filename_entry), FILENAME_MAX);
    gtk_box_pack_start(GTK_BOX(hbox), ConfWidgets.filename_entry, FALSE, TRUE, 2);
    gtk_signal_connect(GTK_OBJECT(ConfWidgets.filename_entry), "changed", GTK_SIGNAL_FUNC(OnFilenameEntry), NULL);
  
    // Display Buttons
    for (i=0; i<CONFIGBUTTONCOUNT; i++) {
        hbox = gtk_hbox_new(FALSE, 0);

        sprintf(buffer, "%d",  PadButtons[currentPad][buttonInfo[i].nr]);
        
        ConfWidgets.button[i] = gtk_button_new_with_label(buttonInfo[i].label);
        gtk_box_pack_start(GTK_BOX(hbox), ConfWidgets.button[i], FALSE, TRUE, 2);
        gtk_fixed_put(GTK_FIXED(fixed), hbox, buttonInfo[i].x, buttonInfo[i].y);
        gtk_object_set_user_data(GTK_OBJECT(ConfWidgets.button[i]), (char *)i);
        gtk_signal_connect(GTK_OBJECT(ConfWidgets.button[i]), "clicked", GTK_SIGNAL_FUNC(OnConfBtn), NULL);
        ConfWidgets.label[i] = gtk_label_new(buffer);
        gtk_box_pack_start(GTK_BOX(hbox), ConfWidgets.label[i], FALSE, TRUE, 2);
    }

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_fixed_put(GTK_FIXED(fixed), hbox, 20, 260);
    ConfWidgets.ok_button = gtk_button_new_with_label("    Ok    ");
    gtk_box_pack_start(GTK_BOX(hbox), ConfWidgets.ok_button, FALSE, TRUE, 2);
    gtk_signal_connect(GTK_OBJECT(ConfWidgets.ok_button), "clicked", GTK_SIGNAL_FUNC(OnConfOk), NULL);
    GTK_WIDGET_SET_FLAGS(ConfWidgets.ok_button, GTK_CAN_DEFAULT);

    ConfWidgets.cancel_button = gtk_button_new_with_label("  Cancel  ");
    gtk_signal_connect(GTK_OBJECT(ConfWidgets.cancel_button), "clicked", GTK_SIGNAL_FUNC(OnConfCancel), NULL);
    gtk_box_pack_start(GTK_BOX(hbox), ConfWidgets.cancel_button, FALSE, TRUE, 2);
    GTK_WIDGET_SET_FLAGS(ConfWidgets.cancel_button, GTK_CAN_DEFAULT);
}

long PADconfigure(void) {
    currentPad = 0;
    CreateConfigWindow();
    gtk_widget_show_all(ConfWidgets.config_window);

    gtk_main();

    return 0;
}


/*---------------------------------------------------------------------*/
/*                          About dialogue stuff                       */
/*---------------------------------------------------------------------*/

static GtkWidget *AboutDlg;

static void OnAboutOk() {
    gtk_widget_hide(AboutDlg);
    gtk_widget_destroy(AboutDlg);
    gtk_main_quit();
}

void PADabout(void) {
    GtkWidget *Label;
    GtkWidget *Ok;
    GtkWidget *Box,*BBox;
    char AboutText[255];

    sprintf (AboutText,"%s %d.%d\n",LibName, revision, build);

    AboutDlg = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_widget_set_usize(AboutDlg, 260, 80);
    gtk_window_set_title(GTK_WINDOW(AboutDlg), "PAD About Dialog");
    gtk_window_set_position(GTK_WINDOW(AboutDlg), GTK_WIN_POS_CENTER);
    gtk_container_set_border_width(GTK_CONTAINER(AboutDlg), 10);

    Box = gtk_vbox_new(0, 0);
    gtk_container_add(GTK_CONTAINER(AboutDlg), Box);
    gtk_widget_show(Box);

    Label = gtk_label_new(AboutText);
    gtk_box_pack_start(GTK_BOX(Box), Label, FALSE, FALSE, 0);
    gtk_widget_show(Label);

    BBox = gtk_hbutton_box_new();
    gtk_box_pack_start(GTK_BOX(Box), BBox, FALSE, FALSE, 0);
    gtk_widget_show(BBox);

    Ok = gtk_button_new_with_label("Ok");
    gtk_signal_connect(GTK_OBJECT(Ok), "clicked", GTK_SIGNAL_FUNC(OnAboutOk), NULL);
    gtk_container_add(GTK_CONTAINER(BBox), Ok);
    GTK_WIDGET_SET_FLAGS(Ok, GTK_CAN_DEFAULT);
    gtk_widget_show(Ok);

    gtk_widget_show(AboutDlg);

    gtk_main();
}

long PADtest(void) {
    int i;
    int f;

    int r=1;

    for (i=0; i<2; i++) {
        if (devicefilename[i][0]) {
            r = 0;
            f = open(devicefilename[i], O_RDONLY);
            if (f == -1) {
                return -1;
            }
            close (f);
        }
    }

    return r;
}
